import numpy as np
from ..components.automls import RunAutoSklearn


automls = {
    "autosklearn": RunAutoSklearn()
}

class CustomAutoML:
    def __init__(self, args, automl):
        self.args = args
        if automl in automls.keys():
            self.automl = automls[automl]
        else:
            raise ValueError("Model {} not defined in model.py".format(automl))
        self.result = None

    def train(self, X_train: np.ndarray, y_train: np.ndarray):
        # Train classifier
        self.automl.fit(X_train, y_train, args=self.args)

    def predict(self):

        result = self.automl.predict()
        return result


if __name__ == "__main__":
    class Arguments:
        def __init__(self):
            self.dataset = "nhanes.csv"
            self.metric = "auroc"

            self.device = "cpu"
            self.cuda = "0"

            self.val_ratio=0.1,
            self.test_raito=0.1,
    args = Arguments()

    tmp = CustomAutoML(args, 'autosklearn')
    tmp.train(None, None)
    print(tmp.predict(None, None))



import os
import traceback
import numpy as np
import logging
import threading
import timeit
import time
import pprint
import torch
from torch import nn
from torch import multiprocessing as mp


from .utils import Buffers, Environment, FileWriter, from_logits, get_batch, compute_baseline_loss, compute_entropy_loss, compute_policy_gradient_loss
from .env import Env, GymWrapper
from .models import CrossInstanceNet, InstanceSpecificNet, LowLevelNet

def train(flags, train_X, train_y, val_X, val_y, test_X, test_y, clf):
    plogger = FileWriter(
        xpid=flags.xpid, xp_args=flags.__dict__, rootdir=flags.savedir
    )

    T = flags.instance_specific_unroll_length
    B = flags.batch_size

    torch.set_num_threads(1)

    env = Env(train_X, train_y, val_X, val_y, clf, flags.metric)
    env = GymWrapper(env, flags, neighbors=1)
    flags.neighbors = min(flags.num_max_neighbors, env.length)
    flags.considered_neighbors = flags.neighbors # reserved for tuning the neighbors. It is fixed now

    env = Env(train_X, train_y, val_X, val_y, clf, flags.metric)
    env = GymWrapper(env, flags)
    test_env = Env(train_X, train_y, test_X, test_y, clf, flags.metric)
    test_env = GymWrapper(test_env, flags)

    flags.num_cross_instance_actions = flags.undersample_ratio * flags.cross_instance_scale // flags.num_instance_specific_actions
    cross_instance_model = CrossInstanceNet(env.observation_space, flags.num_cross_instance_actions)
    cross_instance_model.share_memory()
    cross_instance_buffers = create_cross_instance_buffers(flags, env.observation_space, flags.num_cross_instance_actions)

    instance_specific_model = InstanceSpecificNet(env.observation_space, env.num_instance_specific_actions)
    instance_specific_model.share_memory()
    instance_specific_buffers = create_instance_specific_buffers(flags, env.observation_space, env.num_instance_specific_actions)

    low_level_model = LowLevelNet(env.observation_space, len(flags.ratio_map))
    low_level_model.share_memory()
    low_level_buffers = create_low_level_buffers(flags, env.observation_space)

    actor_processes = []
    ctx = mp.get_context("fork")
    cross_instance_free_queue = ctx.SimpleQueue()
    cross_instance_full_queue = ctx.SimpleQueue()
    instance_specific_free_queue = ctx.SimpleQueue()
    instance_specific_full_queue = ctx.SimpleQueue()
    low_level_free_queue = ctx.SimpleQueue()
    low_level_full_queue = ctx.SimpleQueue()
    data_queue = ctx.Queue()

    for i in range(flags.num_actors):
        actor = ctx.Process(
            target=act,
            args=(
                flags,
                i,
                cross_instance_free_queue,
                cross_instance_full_queue,
                instance_specific_free_queue,
                instance_specific_full_queue,
                low_level_free_queue,
                low_level_full_queue,
                data_queue,
                cross_instance_model,
                cross_instance_buffers,
                instance_specific_model,
                instance_specific_buffers,
                low_level_model,
                low_level_buffers,
                train_X,
                train_y,
                val_X,
                val_y,
                clf
            ),
        )
        actor.start()
        actor_processes.append(actor)

    cross_instance_learner_model = CrossInstanceNet(env.observation_space, flags.num_cross_instance_actions)
    cross_instance_learner_model = cross_instance_learner_model.to(flags.device)
    instance_specific_learner_model = InstanceSpecificNet(env.observation_space, env.num_instance_specific_actions)
    instance_specific_learner_model = instance_specific_learner_model.to(flags.device)
    low_level_learner_model = LowLevelNet(env.observation_space, len(flags.ratio_map), device=flags.device)
    low_level_learner_model = low_level_learner_model.to(flags.device)

    cross_instance_optimizer = torch.optim.Adam(
        cross_instance_learner_model.parameters(),
        lr=flags.learning_rate,
    )

    instance_specific_optimizer = torch.optim.Adam(
        instance_specific_learner_model.parameters(),
        lr=flags.learning_rate,
    )

    low_level_optimizer = torch.optim.Adam(
        low_level_learner_model.parameters(),
        lr=flags.learning_rate,
    )

    def lr_lambda(epoch):
        return 1 - min(epoch * T * B, flags.total_steps) / flags.total_steps

    logger = logging.getLogger("logfile")

    step, cross_instance_stats, stats, low_level_stats = 0, {}, {}, {}

    def cross_instance_batch_and_learn(i):
        """Thread target for the learning process."""
        nonlocal step, cross_instance_stats
        while step < flags.total_steps:
            indices = []

            while len(indices) < flags.batch_size:
                while len(indices) < flags.batch_size and not cross_instance_full_queue.empty():
                    indices.append(cross_instance_full_queue.get())
                time.sleep(1)
                if step >= flags.total_steps:
                    return
            batch = get_batch(
                flags,
                cross_instance_free_queue,
                cross_instance_full_queue,
                cross_instance_buffers,
                indices
            )
            cross_instance_stats = learn(
                flags, cross_instance_model, cross_instance_learner_model, batch, cross_instance_optimizer
            )

    def instance_specific_batch_and_learn(i):
        """Thread target for the learning process."""
        nonlocal step, stats
        while step < flags.total_steps:
            indices = []

            while len(indices) < flags.batch_size:
                while len(indices) < flags.batch_size and not instance_specific_full_queue.empty():
                    indices.append(instance_specific_full_queue.get())
                time.sleep(1)
                if step >= flags.total_steps:
                    return
            batch = get_batch(
                flags,
                instance_specific_free_queue,
                instance_specific_full_queue,
                instance_specific_buffers,
                indices
            )
            stats = learn(
                #flags, model, learner_model, batch, agent_state, optimizer, scheduler
                flags, instance_specific_model, instance_specific_learner_model, batch, instance_specific_optimizer
            )

    def low_level_batch_and_learn(i):
        """Thread target for the learning process."""
        nonlocal step, low_level_stats
        while step < flags.total_steps:
            indices = []

            while len(indices) < flags.batch_size:
                while len(indices) < flags.batch_size and not low_level_full_queue.empty():
                    indices.append(low_level_full_queue.get())
                time.sleep(1)
                if step >= flags.total_steps:
                    return
            batch = get_batch(
                flags,
                low_level_free_queue,
                low_level_full_queue,
                low_level_buffers,
                indices,
            )
            low_level_stats = learn(
                flags, low_level_model, low_level_learner_model, batch, low_level_optimizer
            )

    for m in range(flags.num_buffers):
        cross_instance_free_queue.put(m)
        instance_specific_free_queue.put(m)
        low_level_free_queue.put(m)

    # Cross instance thread
    cross_instance_thread = threading.Thread(
        target=cross_instance_batch_and_learn, name="pre-level-batch-and-learn-%d" % i, args=(i,)
    )
    cross_instance_thread.start()

    # Instance specific thread
    instance_specific_thread = threading.Thread(
        target=instance_specific_batch_and_learn, name="high-level-batch-and-learn-%d" % i, args=(i,)
    )
    instance_specific_thread.start()

    # Low-level thread
    low_level_thread = threading.Thread(
        target=low_level_batch_and_learn, name="low-level-batch-and-learn-%d" % i, args=(i,)
    )
    low_level_thread.start()

    best_val_score = 0
    test_score = 0
    def evaluate():
        nonlocal step, best_val_score, test_score
        while step < flags.total_steps:
            while data_queue.empty():
                time.sleep(1)
                if step >= flags.total_steps:
                    return

            val_score, samples = data_queue.get()
            step += 1
            if val_score > best_val_score:
                best_val_score = val_score
                test_score = test_env.get_reward(samples)
            # print("Current val:", val_score, "best val:", best_val_score, "test:", test_score)

    eval_thread = threading.Thread(
        target=evaluate, name="evaluate-%d" % i
        )
    eval_thread.start()

    timer = timeit.default_timer
    try:
        while step < flags.total_steps:
            start_step = step
            start_time = timer()
            time.sleep(5)

            sps = (step - start_step) / (timer() - start_time)
            if stats.get("episode_retur`ns", None):
                mean_return = (
                    "Return per episode: %.1f. " % stats["mean_episode_return"]
                )
            else:
                mean_rpren = ""
            total_loss = stats.get("total_loss", float("inf"))
            logging.info(
                "Steps %i @ %.1f SPS. Loss %f. %sStats:\n%s\n%s\n%s",
                step,
                sps,
                total_loss,
                "",
                pprint.pformat(cross_instance_stats),
                pprint.pformat(stats),
                pprint.pformat(low_level_stats),
            )
    except KeyboardInterrupt:
        plogger.close()
        return  # Try joining actors then quit.
    else:
        cross_instance_thread.join()
        instance_specific_thread.join()
        low_level_thread.join()
        eval_thread.join()
        logging.info("Learning finished after %d steps.", step)
    finally:
        for _ in range(flags.num_actors):
            cross_instance_free_queue.put(None)
            instance_specific_free_queue.put(None)
            low_level_free_queue.put(None)
        for actor in actor_processes:
            actor.terminate()

    plogger.close()
    return test_score

def learn(
    flags,
    actor_model,
    model,
    batch,
    optimizer,
):
    """Performs a learning (optimization) step."""
    if isinstance(model, LowLevelNet):
        learner_outputs = model(batch, flags.considered_neighbors, flags.device)
        learner_outputs["policy_logits"] = learner_outputs["policy_logits"][:, :, :flags.considered_neighbors*len(flags.ratio_map)]
    else:
        learner_outputs = model(batch, flags.device)

    # Take final value function slice for bootstrapping.
    bootstrap_value = learner_outputs["baseline"][-1]

    # Move from obs[t] -> action[t] to action[t] -> obs[t].
    batch = {key: tensor[1:] for key, tensor in batch.items()}
    learner_outputs = {key: tensor[:-1] for key, tensor in learner_outputs.items()}

    rewards = batch["reward"]
    discounts = (~batch["done"]).float() * flags.discounting

    vtrace_returns = from_logits(
        behavior_policy_logits=batch["policy_logits"],
        target_policy_logits=learner_outputs["policy_logits"],
        actions=batch["action"],
        discounts=discounts,
        rewards=rewards,
        values=learner_outputs["baseline"],
        bootstrap_value=bootstrap_value,
    )

    pg_loss = compute_policy_gradient_loss(
        learner_outputs["policy_logits"],
        batch["action"],
        vtrace_returns.pg_advantages,
    )
    baseline_loss = flags.baseline_cost * compute_baseline_loss(
        vtrace_returns.vs - learner_outputs["baseline"]
    )
    entropy_loss = flags.entropy_cost * compute_entropy_loss(
        learner_outputs["policy_logits"]
    )

    total_loss = pg_loss + baseline_loss + entropy_loss

    episode_returns = batch["episode_return"][batch["done"]]
    stats = {
        "mean_episode_return": torch.mean(episode_returns).item(),
        "total_loss": total_loss.item(),
        "pg_loss": pg_loss.item(),
        "baseline_loss": baseline_loss.item(),
        "entropy_loss": entropy_loss.item(),
    }

    optimizer.zero_grad()
    total_loss.backward()
    nn.utils.clip_grad_norm_(model.parameters(), flags.grad_norm_clipping)
    optimizer.step()

    actor_model.load_state_dict(model.state_dict())
    return stats

def act(
    flags,
    actor_index: int,
    cross_instance_free_queue: mp.SimpleQueue,
    cross_instance_full_queue: mp.SimpleQueue,
    instance_specific_free_queue: mp.SimpleQueue,
    instance_specific_full_queue: mp.SimpleQueue,
    low_level_free_queue: mp.SimpleQueue,
    low_level_full_queue: mp.SimpleQueue,
    data_queue: mp.Queue,
    cross_instance_model,
    cross_instance_buffers,
    instance_specific_model,
    instance_specific_buffers,
    low_level_model,
    low_level_buffers,
    train_X,
    train_y,
    val_X,
    val_y,
    clf
):
    try:
        logging.info("Actor %i started.", actor_index)

        torch.set_num_threads(1)

        env = Env(train_X, train_y, val_X, val_y, clf, flags.metric)
        gym_env = GymWrapper(env, flags)
        seed = actor_index ^ int.from_bytes(os.urandom(4), byteorder="little")
        torch.manual_seed(seed)
        np.random.seed(seed)
        gym_env.seed(seed)
        env = Environment(gym_env)
        env_output = env.initial()

        cross_instance_agent_output = cross_instance_model(env_output)
        instance_specific_agent_output = instance_specific_model(env_output)
        low_level_agent_output = low_level_model(env_output, flags.considered_neighbors)

        # Some preperation
        cross_instance_env_output_keys = list(env_output.keys())
        cross_instance_env_output_keys.remove("neighbor_features")

        instance_specific_env_output_keys = list(env_output.keys())
        instance_specific_env_output_keys.remove("neighbor_features")
        instance_specific_action = 0

        # tmp buffer to store historical pre level episodes
        tmp_cross_instance_buf = {}
        for key in cross_instance_env_output_keys:
            tmp_cross_instance_buf[key] = []
        for key in cross_instance_agent_output:
            tmp_cross_instance_buf[key] = []

        # Write the inital rollout (pre-level).
        for key in cross_instance_env_output_keys:
            tmp_cross_instance_buf[key].append(env_output[key])
        for key in cross_instance_agent_output:
            tmp_cross_instance_buf[key].append(cross_instance_agent_output[key])


        # tmp buffer to store historical low level episodes
        tmp_low_level_buf = {}
        for key in env_output:
            tmp_low_level_buf[key] = []
        for key in low_level_agent_output:
            tmp_low_level_buf[key] = []

        # tmp splitted low level buffer to track the reward
        tmp_splitted_low_level_buf = {}
        for key in env_output:
            tmp_splitted_low_level_buf[key] = []
        for key in low_level_agent_output:
            tmp_splitted_low_level_buf[key] = []

        # Write the inital rollout (low-level).
        for key in env_output:
            tmp_low_level_buf[key].append(env_output[key])
        for key in low_level_agent_output:
            tmp_low_level_buf[key].append(low_level_agent_output[key])

        act_done = False
        while not act_done:
            instance_specific_index = instance_specific_free_queue.get()
            if instance_specific_index is None:
                act_done = True
                break

            # Write old rollout end (high-level).
            for key in instance_specific_env_output_keys:
                instance_specific_buffers[key][instance_specific_index][0, ...] = env_output[key]
            for key in instance_specific_agent_output:
                instance_specific_buffers[key][instance_specific_index][0, ...] = instance_specific_agent_output[key]

            # Do new rollout.
            for t in range(flags.instance_specific_unroll_length):

                instance_specific_scale = cross_instance_agent_output["action"].item()
                with torch.no_grad():
                    instance_specific_agent_output = instance_specific_model(env_output)
                instance_specific_action = instance_specific_agent_output["action"].item()
                instance_specific_action *= instance_specific_scale
                #instance_specific_action = 50 # hard-code

                # Create a new entry in the tmp low-level buffer
                for key in tmp_splitted_low_level_buf:
                    tmp_splitted_low_level_buf[key].append([])
                to_next = False
                while not to_next:
                    if instance_specific_action > 0:
                        with torch.no_grad():
                            low_level_agent_output = low_level_model(env_output, flags.considered_neighbors)
                        low_level_action = low_level_agent_output["action"].item()
                        dst_index = low_level_action // len(flags.ratio_map)
                        ratio = flags.ratio_map[low_level_action % len(flags.ratio_map)] 
                    else:
                        dst_index = -1
                        ratio = -1
                    to_next = True if instance_specific_action <= 1 else False

                    low_level_action = {
                        "dst_index": dst_index,
                        "ratio": ratio,
                        "next": to_next
                    }
                    env_output, info = env.step(low_level_action)
                    for key in env_output:
                        tmp_splitted_low_level_buf[key][-1].append(env_output[key])
                    for key in low_level_agent_output:
                        tmp_splitted_low_level_buf[key][-1].append(low_level_agent_output[key])
                    instance_specific_action -= 1

                if env_output["done"]:

                    for key in cross_instance_env_output_keys:
                        tmp_cross_instance_buf[key].append(env_output[key])
                    for key in cross_instance_agent_output:
                        tmp_cross_instance_buf[key].append(cross_instance_agent_output[key])

                    while len(tmp_cross_instance_buf["reward"]) > flags.cross_instance_unroll_length+1:
                        cross_instance_index = cross_instance_free_queue.get()
                        if cross_instance_index is None:
                            act_done = True
                            break

                        for key in tmp_cross_instance_buf:
                            for pre_t in range(flags.cross_instance_unroll_length+1):
                                cross_instance_buffers[key][cross_instance_index][pre_t, ...] = tmp_cross_instance_buf[key][pre_t]
                            tmp_cross_instance_buf[key] = tmp_cross_instance_buf[key][flags.cross_instance_unroll_length:]
                        cross_instance_full_queue.put(cross_instance_index)

                    cross_instance_agent_output = cross_instance_model(env_output)

                    # Assign rewards to low-level episodes, and sample
                    idx = [i for i in range(len(tmp_splitted_low_level_buf["reward"]))]
                    sampled_idx = np.random.choice(idx, size=10)
                    for key in tmp_splitted_low_level_buf:
                        tmp_splitted_low_level_buf[key] = [tmp_splitted_low_level_buf[key][i] for i in sampled_idx]

                    # Assign rewards to low-level episodes, and sample
                    for episode_id in range(len(tmp_splitted_low_level_buf["reward"])):
                        if len(tmp_splitted_low_level_buf["reward"][episode_id]) > 0:
                            tmp_splitted_low_level_buf["reward"][episode_id][-1] = env_output["reward"]
                            tmp_splitted_low_level_buf["done"][episode_id][-1] = env_output["done"]

                    # Flatten and put them to tmp_low_level_buf
                    for key in tmp_low_level_buf:
                        for episode_data in tmp_splitted_low_level_buf[key]:
                            tmp_low_level_buf[key].extend(episode_data)
                        tmp_splitted_low_level_buf[key] = []

                    # Put the data into the buffer
                    while len(tmp_low_level_buf["reward"]) > flags.low_level_unroll_length+1:
                        low_level_index = low_level_free_queue.get()
                        if low_level_index is None:
                            act_done = True
                            break

                        for key in tmp_low_level_buf:
                            for low_t in range(flags.low_level_unroll_length+1):
                                low_level_buffers[key][low_level_index][low_t, ...] = tmp_low_level_buf[key][low_t]
                            tmp_low_level_buf[key] = tmp_low_level_buf[key][flags.low_level_unroll_length:]
                        low_level_full_queue.put(low_level_index)

                    data_queue.put((env_output["episode_return"].numpy().flatten()[0], info["samples"]))

                for key in instance_specific_env_output_keys:
                    instance_specific_buffers[key][instance_specific_index][t + 1, ...] = env_output[key]
                for key in instance_specific_agent_output:
                    instance_specific_buffers[key][instance_specific_index][t + 1, ...] = instance_specific_agent_output[key]

            instance_specific_full_queue.put(instance_specific_index)

    except KeyboardInterrupt:
        pass  # Return silently.
    except Exception as e:
        logging.error("Exception in worker process %i", actor_index)
        traceback.print_exc()
        print()
        raise e

def create_cross_instance_buffers(flags, observation_space, num_actions) -> Buffers:
    T = flags.cross_instance_unroll_length
    specs = dict(
        src_features=dict(size=(T + 1, *observation_space["src_features"].shape), dtype=torch.float32),
        reward=dict(size=(T + 1,), dtype=torch.float32),
        done=dict(size=(T + 1,), dtype=torch.bool),
        episode_return=dict(size=(T + 1,), dtype=torch.float32),
        episode_step=dict(size=(T + 1,), dtype=torch.int32),
        policy_logits=dict(size=(T + 1, num_actions), dtype=torch.float32),
        baseline=dict(size=(T + 1,), dtype=torch.float32),
        action=dict(size=(T + 1,), dtype=torch.int64),
    )
    buffers: Buffers = {key: [] for key in specs}
    for _ in range(flags.num_buffers):
        for key in buffers:
            buffers[key].append(torch.empty(**specs[key]).share_memory_())
    return buffers

def create_instance_specific_buffers(flags, observation_space, num_actions) -> Buffers:
    T = flags.instance_specific_unroll_length
    specs = dict(
        src_features=dict(size=(T + 1, *observation_space["src_features"].shape), dtype=torch.float32),
        reward=dict(size=(T + 1,), dtype=torch.float32),
        done=dict(size=(T + 1,), dtype=torch.bool),
        episode_return=dict(size=(T + 1,), dtype=torch.float32),
        episode_step=dict(size=(T + 1,), dtype=torch.int32),
        policy_logits=dict(size=(T + 1, num_actions), dtype=torch.float32),
        baseline=dict(size=(T + 1,), dtype=torch.float32),
        action=dict(size=(T + 1,), dtype=torch.int64),
    )
    buffers: Buffers = {key: [] for key in specs}
    for _ in range(flags.num_buffers):
        for key in buffers:
            buffers[key].append(torch.empty(**specs[key]).share_memory_())
    return buffers

def create_low_level_buffers(flags, observation_space) -> Buffers:
    T = flags.low_level_unroll_length
    specs = dict(
        src_features=dict(size=(T + 1, *observation_space["src_features"].shape), dtype=torch.float32),
        neighbor_features=dict(size=(T + 1, *observation_space["neighbor_features"].shape), dtype=torch.float32),
        reward=dict(size=(T + 1,), dtype=torch.float32),
        done=dict(size=(T + 1,), dtype=torch.bool),
        episode_return=dict(size=(T + 1,), dtype=torch.float32),
        episode_step=dict(size=(T + 1,), dtype=torch.int32),
        policy_logits=dict(size=(T + 1, observation_space["neighbor_features"].shape[0]*len(flags.ratio_map)), dtype=torch.float32),
        baseline=dict(size=(T + 1,), dtype=torch.float32),
        action=dict(size=(T + 1,), dtype=torch.int64),
    )
    buffers: Buffers = {key: [] for key in specs}
    for _ in range(flags.num_buffers):
        for key in buffers:
            buffers[key].append(torch.empty(**specs[key]).share_memory_())
    return buffers



import numpy as np



# from .hybrids.runspe import RunSPE
# from .hybrids.runmesa import RunMESA
from ..components.hybrids import RunAutoSmote
from ..processing.utils import ArgsNamespace


hbds = {
    "autosmote": RunAutoSmote(),
    # "spe": RunSPE(),
    # "mesa": RunMESA(),
}

class CustomHybrid:
    def __init__(self, args: ArgsNamespace, pipe=None):
        self.args = args
        imp, hbd = pipe
        if hbd in hbds.keys():
            self.hbd = hbds[hbd]
            if not hasattr(self.hbd, "data_folder_path"):
                raise Exception("Model {} does not have data_folder_path attribute".format(self.hbd))
            setattr(self.hbd, "data_folder_path", self.args.path)
        else:
            raise Exception("Model {} not defined in model.py".format(hbd))
        self.imp = imp
        self.result = None

    def train(self, X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray = None, y_test: np.ndarray = None):
        # Train classifier

        self.hbd.fit(X_train, y_train, X_test, y_test, args=self.args, imp=self.imp)

        # Sample for asynchronize code
        # async def main():
        #     async with aiohttp.ClientSession() as session:
        #         async with session.get('https://api.example.com') as response:
        #             print("Requested!")
        #             return response
        #
        # asyncio.run(main())

        # async def get_responses(n):
        #     async with aiohttp.ClientSession() as session:
        #         responses = await asyncio.gather(session.post(url))
        #         assert all(r.status == 200 for r in responses)
        #         return [await r.text() for r in responses]

    def predict(self, X_test: np.ndarray, y_test: np.ndarray):

        result = self.hbd.predict(X_test, y_test)
        self.result = result
        return result


if __name__ == "__main__":
    class Arguments:
        def __init__(self):
            self.dataset = "nhanes.csv"
            self.metric = "auroc"

            self.device = "cpu"
            self.cuda = "0"

            self.val_ratio=0.1,
            self.test_raito=0.1,
    args = Arguments()

    tmp = CustomHybrid(args, ['median', 'autosmote'])
    tmp.train(None, None)



import os
from typing import Dict

import gensim
from gensim.models.doc2vec import Doc2Vec, TaggedDocument
from sklearn.metrics.pairwise import cosine_similarity
from sys import exit

from .filter_utils import read_txt_files_to_corpus, count_top_models, get_query, find_best_pipes, get_accuracy


# query = "This dataset is provided by the National Institute of Diabetes and Digestive and Kidney Diseases and focuses on predicting the onset of diabetes in a population of Pima Indian women. It contains 768 instances, each with 8 attributes, including factors such as plasma glucose concentration, diastolic blood pressure, and body mass index. The dataset's class distribution is imbalanced, with 268 positive cases and 500 negative cases. The missing ratio is 10%."

# data = "Resamplers"

class RunDoc2Vec:
    def __init__(self, documents, topn=3):
        self.documents = documents
        self.topn = topn

        # Tagging the documents (Each document must have a unique tag)
        self.tagged_data = [TaggedDocument(words=self.preprocess(doc), tags=[str(i)]) for i, doc in enumerate(documents)]
        self.model = Doc2Vec(vector_size=100, window=5, min_count=2, workers=4, epochs=40)

    def preprocess(self, doc):
        return gensim.utils.simple_preprocess(doc)

    def train(self, corpus, query):
        self.model.build_vocab(self.tagged_data)
        self.model.train(self.tagged_data, total_examples=self.model.corpus_count, epochs=self.model.epochs)
        # Function to infer a vector for a new document
        def get_doc_vector(doc):
            preprocessed_doc = self.preprocess(doc)
            return self.model.infer_vector(preprocessed_doc)

        # Compare similarity between two documents
        def compare_documents(doc1, doc2):
            vec1 = get_doc_vector(doc1)
            vec2 = get_doc_vector(doc2)

            # Calculate cosine similarity
            similarity = cosine_similarity([vec1], [vec2])
            return similarity[0][0]

        # Example: Comparing document similarity
        doc_names = list(corpus.keys())
        final_score = []
        for i, doc in enumerate(self.documents):
            similarity_score = compare_documents(doc, query)
            final_score.append([doc_names[i], similarity_score])

        final_score = sorted(final_score, key=lambda x: x[1], reverse=True)

        return [row[0] for row in final_score][:self.topn]

# model_folder_path = os.path.join("..", 'data', 'models', data)
#
# corpus = read_txt_files_to_corpus(model_folder_path)
#
# # Example training data
# documents = corpus.values()
#
# # Preprocess the documents (optional, depending on the use case)
# def preprocess(doc):
#     return gensim.utils.simple_preprocess(doc)
#
# # Tagging the documents (Each document must have a unique tag)
# tagged_data = [TaggedDocument(words=preprocess(doc), tags=[str(i)]) for i, doc in enumerate(documents)]
#
# # Initialize the Doc2Vec model
# model = Doc2Vec(vector_size=100, window=5, min_count=2, workers=4, epochs=40)
#
# # Build vocabulary from the corpus
# model.build_vocab(tagged_data)
#
# # Train the model
# model.train(tagged_data, total_examples=model.corpus_count, epochs=model.epochs)
#
#
# # Function to infer a vector for a new document
# def get_doc_vector(doc):
#     preprocessed_doc = preprocess(doc)
#     return model.infer_vector(preprocessed_doc)
#
#
# # Compare similarity between two documents
# def compare_documents(doc1, doc2):
#     vec1 = get_doc_vector(doc1)
#     vec2 = get_doc_vector(doc2)
#
#     # Calculate cosine similarity
#     similarity = cosine_similarity([vec1], [vec2])
#     return similarity[0][0]
#
#
# # Example: Comparing document similarity
# doc_names = list(corpus.keys())
# final_score = []
# for i, doc in enumerate(documents):
#     similarity_score = compare_documents(doc, query)
#     final_score.append([doc_names[i], similarity_score])
#
# final_score = sorted(final_score, key=lambda x: x[1], reverse=True)
# for key, value in final_score:
#     print(key, value)


def get_ranking(datasets, topn=3):
    """ Get the ranking from model description and dataset description """
    LLM_result = {}
    for dataset in datasets:
        dataset_result = {}
        LLM_result[dataset] = dataset_result
        query = get_query(dataset)
        for strategy in strategies:
            method_result = {}
            for model_type in model_types:
                model_folder_path = os.path.join("../../../..", 'data', 'models', 'dps', strategy, model_type)
                index_folder_path = os.path.join("../../../..", 'data', 'processed', 'dps', strategy, model_type)
                corpus = read_txt_files_to_corpus(model_folder_path)
                results = {}
                ranker = RunDoc2Vec(corpus.values())
                model_result = ranker.train(corpus, query)
                method_result[model_type] = model_result
            dataset_result[strategy] = method_result

    return LLM_result


if __name__ == "__main__":
    # methods = ["combined", "sequential", "interactive"]
    # strategies = ["combined", "sequential"]
    # strategies = ["combined"]
    strategies = ["sequential"]

    datasets = ["Pima", "Ljubljana", "Wisconsin", "NHANES"]
    # datasets = ["NHANES"]

    # query = "This dataset is provided by the National Institute of Diabetes and Digestive and Kidney Diseases and focuses on predicting the onset of diabetes in a population of Pima Indian women. It contains 768 instances, each with 8 attributes, including factors such as plasma glucose concentration, diastolic blood pressure, and body mass index. The dataset's class distribution is imbalanced, with 268 positive cases and 500 negative cases. The missing value ratio is 10\% across all features."
    # query = "This dataset is provided by the National Institute of Diabetes and Digestive and Kidney Diseases and focuses on predicting the onset of diabetes in a population of Pima Indian women. It contains 768 instances, each with 8 attributes, including factors such as plasma glucose concentration, diastolic blood pressure, and body mass index. The dataset's class distribution is imbalanced, with 268 positive cases and 500 negative cases. The missing ratio is 10%."

    model_types = ["imputer", "resampler", "classifier"]
    metrics = ["auroc", "macro_f1"]
    # metrics = ["auroc"]

    # ground_truth Structure: {metric: {dataset: {model_type: top count}}}
    ground_truth = find_best_pipes(datasets, metrics)

    # formatted_data = json.dumps(ground_truth, indent=4)
    # pprint(formatted_data)

    # LLM_result Structure: {dataset: {strategy: {model_type: model ranking}}}
    LLM_result = get_ranking(datasets)

    LLM_accuracy = {strategy:
                        {dataset:
                             {metric: {} for metric in metrics} for dataset in datasets} for strategy in strategies}

    for metric in metrics:
        for strategy in strategies:
            for dataset in ground_truth[metric]:
                LLM_accuracy[strategy][dataset][metric] = get_accuracy(ground_truth[metric][dataset], LLM_result[dataset][strategy])

    print(LLM_accuracy)



from ChatGPTprompt import PromptEngineering
import os
from ...processing.utils import DataLoader

model_abbreviation = {
    "lr" : "Logistic Regression",
    "mlp": "Multi-layer Perceptron",
    "ada": "Adaboost",
    "svm": "Support Vector Machine",
    "median": "Median Value Imputation",
    "knn": "kNN Imputation",
    "ii": "Iterative Imputation",
    "gain": "GAIN Imputation",
    "MIRACLE": "MIRACLE Imputation",
    "MIWAE": "MIWAE Imputation",
    'rus': "Random Under Sampler",
    'ros': "Random Over Sampler",
    'smote': "SMOTE",
    'mwmote': "MWMOTE",
}

data_loader = DataLoader()
origin_folder = data_loader.get_models_dp_origins_folder()
api_key_path = os.path.join(data_loader.get_raw_data_folder(), "openai_key.txt")
with open(api_key_path, 'r') as f:
    api_key = f.read()[:-1]

# chat_prompt = PromptEngineering(api_key=api_key)



from sklearn.base import BaseEstimator
import numpy as np
import smote_variants as sv
import logging

logging.getLogger(sv.__name__).setLevel(logging.CRITICAL)


class MWMOTE(BaseEstimator):
    """MWMOTE"""

    def __init__(self, sampling_strategy=1, random_state=0, info_method=None):
        self.X = None
        self.y = None
        self.random_state = random_state
        self.info_method = info_method
        self.sampling_strategy = sampling_strategy
        self.resampler = sv.MWMOTE(proportion=self.sampling_strategy, random_state=42)

    def fit_resample(self, X: np.ndarray, y: np.ndarray):
        """Fit and transform X and y for under sampling"""
        X_samp, y_samp = self.resampler.sample(X, y)

        return X_samp, y_samp



import os
import logging
from abc import ABC, abstractmethod
from flask import Flask, request, jsonify
import pandas as pd
import numpy as np
from types import SimpleNamespace

class Arguments:
    def __init__(self, dictionary):
        for k, v in dictionary.items():
            setattr(self, k, v)

class BaseModelAPI(ABC):
    def __init__(self, import_name):
        self.app = Flask(import_name)
        self.params = {}
        self.result = None

        # Register routes
        self.app.add_url_rule('/set', view_func=self.set_params, methods=['POST'])
        self.app.add_url_rule('/train/<dataset_name>', view_func=self.train, methods=['GET'])
        self.app.add_url_rule('/health', view_func=self.health, methods=['GET'])
        self.app.add_url_rule('/predict', view_func=self.get_result, methods=['POST'])
        self.app.add_url_rule('/hyperparameters', view_func=self.get_hyperparameters(), methods=['GET'])


    def get_hyperparameters(self):
        return jsonify(self.get_hyperparameter_search_space())

    @abstractmethod
    def get_hyperparameter_search_space(self) -> dict:
        # Returns a dictionary that defines the hyperparameters and their ranges/types.
        pass

    def dict_to_namespace(self):
        # Parse dict object to python class attributes like object
        def recurse(d):
            ns = {}
            for k, v in d.items():
                if isinstance(v, dict) and "default" in v:
                    ns[k] = v["default"]
                elif isinstance(v, dict):
                    ns[k] = recurse(v)
                else:
                    ns[k] = v  # fallback
            return SimpleNamespace(**ns)
        return recurse(self.params)

    def health(self):
        return "OK", 200

    def set_params(self):
        """Set training parameters"""
        def recursive_update(d, u):
            for k, v in u.items():
                if isinstance(v, dict) and isinstance(d.get(k), dict):
                    recursive_update(d[k], v)
                else:
                    d[k] = v

        data = request.get_json()

        # TODO delete
        print(data)

        recursive_update(self.params, data)

        if 'metric' not in self.params:
        # if 'metric' not in self.params
                raise Exception("data not complete, need to include metric")
        return jsonify(self.params), 201

    @abstractmethod
    def fit_train(self, args, X_train, y_train, X_test, y_test):
        """Control the actual model training process"""
        pass

    def train(self, dataset_name: str):
        """Load data, run training, and return result"""
        args = Arguments(self.params)

        X_train = pd.read_csv(os.path.join("/data/interim", args.data_names[0])).to_numpy()
        y_train = pd.read_csv(os.path.join("/data/interim", args.data_names[1])).to_numpy().ravel()
        X_test = pd.read_csv(os.path.join("/data/interim", args.data_names[2])).to_numpy()
        y_test = pd.read_csv(os.path.join("/data/interim", args.data_names[3])).to_numpy().ravel()
        logging.info("loading finished")

        self.result = self.fit_train(args, X_train, y_train, X_test, y_test)
        logging.info("finished training")
        return {}, 200
        # return jsonify({"result": self.result}), 200

    def get_result(self):
        return jsonify({"result": self.result}), 200

    def run(self, host='0.0.0.0', port=8080, debug=True):
        self.app.run(host=host, port=port, debug=debug)



from collections import deque
import json


class TreeNode:
    def __init__(self, name, value, node_type, selected_models=None, left=None, right=None):
        # if node_type not in ["root", "imp", "automl", "hbd", "rsp", "clf"]:
        #     raise ValueError("Node type {} not supported in TreeNode".format(node_type))
        self.node_name = name
        self.node_type = node_type
        self.left = left
        self.right = right
        self.max_leaf_value = value  # Initialize max_leaf_value with node value
        self.max_value_pipe = []
        self.index = {}
        self.available_models = selected_models if selected_models else []

class BinaryTree:
    def __init__(self):
        self.root = TreeNode("root", -1, "root")

    def insert(self, name, value, node_type, parent_type, selected_models=None):
        node = TreeNode(name, value, node_type, selected_models)

        self._search_insert(parent_type, node, self.root)

    def _search_insert(self, parent_type, child, parent):
        if parent is None:
            return
        if parent.node_type != parent_type:
            self._search_insert(parent_type, child, parent.left)
            self._search_insert(parent_type, child, parent.right)
        else:
            self._insert(parent, child)

    def _insert(self, parent: TreeNode, child: TreeNode):
        if parent.left is None:
            parent.left = child
        elif parent.right is None:
            if parent.left.max_leaf_value >= child.max_leaf_value:
                parent.right = child
            else:
                parent.right = parent.left
                parent.left = child
        else:
            self.print_tree()
            raise ValueError("Tree node is full, no more insert")

    def _search_recursive(self, node, node_type, max_value=None, node_name=None):
        """ Search for node_type in BinaryTree
            if max_value is not None, replace the max_leaf_value for all the element of the path

        Return
        ----------
        The path found
        """
        if node is None:
            return None

        if node.node_type == node_type:
            if max_value is not None:
                node.max_leaf_value = max_value
                node.node_name = node_name
            return [node.node_name]

        for child in [node.left, node.right]:
            if child:
                path = self._search_recursive(child, node_type, max_value, node_name)
                if path:
                    if max_value is not None:
                        node.max_leaf_value = max(node.left.max_leaf_value if node.left else -1,
                                                  node.right.max_leaf_value if node.right else -1)
                        self._ensure_order(node)
                    return [node.node_name] + path
        return None

    @staticmethod
    def _ensure_order(node):
        """Ensure that left child always has the largest max_leaf_value."""
        if node.left and node.right and node.left.max_leaf_value < node.right.max_leaf_value:
            node.left, node.right = node.right, node.left

    def replace(self, node_type, max_value, node_name):
        """ Replace the max_leaf_value for all the element of the path """
        # node = self._search(node_type, max_value)
        node = self._search_recursive(self.root, node_type, max_value, node_name=node_name)
        if node is None:
            return None
        self.root.max_leaf_value = max_value

    def _search_node(self, node: TreeNode, target_type) -> TreeNode:
        """Search for a given node type downward from a node in a tree"""
        if node is None:
            return False
        if node.node_type == target_type:
            return node
        else:
            left = self._search_node(node.left, target_type)
            if left:
                return left

            right = self._search_node(node.right, target_type)
            if right:
                return right


    def update_pipe(self, pipe, value):
        """ check if sub-branch of the  BinaryTree need to be updated or not

        Return
        ----------
        Ture : if it updated the BinaryTree
        False : if it didn't update the BinaryTree
        """
        if len(pipe) == 1:
            target_node = self._search_node(self.root, "automl")
            if target_node.max_leaf_value < value:
                self.replace("automl", value, pipe[0])
                return True
            else:
                return False

        elif len(pipe) == 2:
            target_node = self._search_node(self.root, "hbd")
            if target_node.max_leaf_value < value:
                self.replace("imp", value, pipe[0])
                self.replace("hbd", value, pipe[1])
                return True
            else:
                return False

        elif len(pipe) == 3:
            target_node = self._search_node(self.root, "clf")
            if target_node.max_leaf_value < value:
                self.replace("imp", value, pipe[0])
                self.replace("rsp", value, pipe[1])
                self.replace("clf", value, pipe[2])
                return True
            else:
                return False

    def best_pipe(self):
        node = self.root.left
        pipe = []
        while node is not None:
            pipe.append(node.node_name)
            node = node.left

        return pipe

    def sub_best_pipe(self, node_type="automl"):
        """ Search for the best pipe built with node_type """
        node = self._search_node(self.root, target_type=node_type)
        pipe = []
        while node is not None:
            pipe.append(node.node_name)
            node = node.left
        return pipe

    def build_pipe(self, node_type):
        """Build the pipeline with the last element of the pipeline as node_type"""
        if node_type not in ["clf", "hbd", "automl"]:
            raise ValueError(
                "Please select the last element of the pipeline as the node_type, {} is not".format(node_type))
        else:
            result = self._search_recursive(self.root, node_type=node_type)
            return result[1:]

    def print_tree(self):
        print("----------------")
        self._print_tree_recursive(self.root, 0)
        print("----------------")

    def _print_tree_recursive(self, node, level):
        if node is not None:
            self._print_tree_recursive(node.right, level + 1)
            print("     " * level + str(node.node_name) + " " + str(node.node_type) + " (" + str(node.max_leaf_value) + ")" + str(node.available_models))
            self._print_tree_recursive(node.left, level + 1)

    def serialize(self):
        """Serialize the tree to a JSON string"""
        def serialize_node(node):
            if node is None:
                return None
            return {
                'name': node.node_name,
                'value': node.max_leaf_value,
                'type': node.node_type,
                'selected_models': node.available_models,
                'left': serialize_node(node.left),
                'right': serialize_node(node.right)
            }

        return json.dumps(serialize_node(self.root), indent=4)

    def save_to_file(self, filename):
        """Save the serialized tree to a JSON file."""
        serialized_tree = self.serialize()
        with open(filename, 'w') as file:
            file.write(serialized_tree)

    @classmethod
    def load_from_file(cls, filename):
        """Load a tree from a JSON file and deserialize it."""
        with open(filename, 'r') as file:
            serialized_tree = file.read()
        return cls.deserialize(serialized_tree)

    @classmethod
    def deserialize(cls, data):
        """Deserialize the JSON string to reconstruct the tree."""
        def deserialize_node(node_data):
            if node_data is None:
                return None
            node = TreeNode(node_data['name'], node_data['value'], node_data['type'], node_data['selected_models'])
            node.left = deserialize_node(node_data['left'])
            node.right = deserialize_node(node_data['right'])
            return node

        data = json.loads(data)
        root = deserialize_node(data)
        tree = cls()
        tree.root = root
        return tree


if __name__ == "__main__":
    tree = BinaryTree()
    tree.insert("mean", 0, "imp", "root", selected_models=['lr', 'mean'])
    tree.insert("autosklearn", 0, "automl", "root")
    tree.insert("s2sl", 0, "rsp", "imp")
    tree.insert("autosmote", 0, "hbd", "imp")
    tree.insert("svm", 0, "clf", "rsp")
    tree.print_tree()
    tree.save_to_file("test.json")

    load_tree = BinaryTree.load_from_file("test.json")
    load_tree.print_tree()




import os

from ...processing.utils import DataLoader
from .filter_utils import read_txt_files_to_corpus
from .runDoc2vec import RunDoc2Vec


class ModelFiltering:
    """Run model filtering before running in AutoML system"""
    def __init__(self,
                 data_description,
                 datafolder_path,                  # The root path of the data folder
                 method='lda',
                 strategy='combined',
                 ):
        self.data_description = data_description
        self.method = method
        self.strategy = strategy
        self.datafolder_path=datafolder_path

        # self.rankers = {"lda": }

    def get_topn(self, model_type, topn=3):
        """ Get the ranking from model description and dataset description """
        from LDAword2vec import LDAword2vec

        model_folder_path = os.path.join(self.datafolder_path, 'models', 'dps', self.strategy, model_type)
        index_folder_path = os.path.join(self.datafolder_path, 'processed', 'dps', self.strategy, model_type)
        corpus = read_txt_files_to_corpus(model_folder_path)

        ranker = LDAword2vec(index_folder_path, rerun=True)
        model_result = ranker.search(corpus=corpus, query=self.data_description, request_doc_num=topn)

        return model_result

    # def get_topn(self, model_type: str, request_doc_num=3):
    #     data_location = DataLoader()
    #     corpus = read_txt_files_to_corpus(os.path.join(data_location.get_models_dp_folder(), model_type))
    #     print(corpus)
    #
    #     ranker = LDAword2vec(data_location.get_interim_data_folder(), rerun=True)
    #     models = ranker.search(corpus=corpus, query=self.data_description, request_doc_num=request_doc_num)
    #     return models



from flask import Flask, jsonify, request
import os
import re
import shutil
import pandas as pd
from sklearn.model_selection import train_test_split
import h2o
from h2o.sklearn import H2OAutoMLClassifier

app = Flask(__name__)
params = {}

@app.route('/set', methods=['post'])
def set_params():
    """ Set training parameters """
    data = request.get_json()
    global params
    for key, value in data.items():
        params[key] = value
    if 'metric' not in params or 'dataset' not in params or 'target' not in params:
        raise Exception("data not complete, need to include metric, dataset, and target")

    return jsonify(params), 201

@app.route('/results/<dataset_name>', methods=['GET'])
def get_result(dataset_name: str):
    """ Train the auto-sklearn and return the result in json format """
    h2o.init(ip="localhost", port=54323)
    global params
    # shutil.rmtree('/tmp', ignore_errors=True)
    null_values = ['', ' ']
    data_path = os.path.join("/data/raw", params['dataset'])
    if not os.path.isfile(data_path):
        return jsonify({"result": "file not exist"}, 200)

    file_type = re.search("[^\.]*$", data_path).group()
    if file_type == "csv":
        df = pd.read_csv(data_path, na_values=null_values)
    elif file_type in ["xls,", "xlsx,", "xlsm,", "xlsb,", "odf,", "ods,", "odt"]:
        df = pd.read_excel(data_path, na_values=null_values)
    else:
        return jsonify({"error": "file type of {} not supported".format(dataset_name)}), 400

    headers = list(df.columns.values)
    target = params["target"]
    headers.remove(target)

    X = df[headers]
    y = df[[target]]

    # The AutoML part
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, stratify=y,test_size=0.1
    )


    if params["metric"] == "auroc":
        model = H2OAutoMLClassifier(max_models=10, seed=42, sort_metric='auc')
    else:
        raise Exception("Evaluation metric {} is not yet supported".format(params["metric"]))


    model.fit(X_train, y_train)
    from sklearn.metrics import  roc_auc_score
    if params["metric"] == "auroc":
        y_proba = model.predict_proba(X_test)
        result = roc_auc_score(y_test, y_proba)
    else:
        raise Exception("Evaluation metric {} is not yet supported".format(params["metric"]))
    return jsonify({"result": result}), 200

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8081, debug=True)


from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier

name2f = {
    "knn": KNeighborsClassifier(),
    "svm": LinearSVC(random_state=42),
    "dt": DecisionTreeClassifier(),
    "ada": AdaBoostClassifier(random_state=42),
    "lr": LogisticRegression(random_state=42),
    "mlp": MLPClassifier(random_state=42, max_iter=1000),
}

def get_clf(name):
    return name2f[name]

